
from os import environ
from pathlib import Path

import click
import pydantic


_addon_dirs = ([True,
                lambda: Path('/Applications/World of Warcraft/Interface/AddOns')],
               ['Public' in environ,
                lambda: Path(environ['Public'],
                             'Games\\World of Warcraft\\Interface\\AddOns')],
               ['ProgramFiles(x86)' in environ,
                lambda: Path(environ['ProgramFiles(x86)'],
                             'World of Warcraft\\Interface\\AddOns')],
               ['ProgramFiles' in environ,
                lambda: Path(environ['ProgramFiles'],
                             'World of Warcraft\\Interface\\AddOns')],)


class Config(pydantic.BaseSettings):

    addon_dir: Path
    config_dir: Path = click.get_app_dir('instawow')
    db_name: str = 'db.sqlite'

    def _process_values(self, values_dict):
        values = super()._process_values(values_dict)
        for k, v in values.items():
            if isinstance(v, Path):
                values[k] = v.expanduser().resolve()
        if not values['addon_dir'] or not values['addon_dir'].is_dir():
            raise ValueError
        return values

    def create_dirs(self):
        """Create the requisite application directories."""
        self.config_dir.mkdir(exist_ok=True)
        return self


class _UserConfigMeta(Config.__class__):

    @property
    def default_addon_dir(cls):
        return next((v() for k, v in _addon_dirs if k and v().is_dir()),
                    None)


class UserConfig(Config, metaclass=_UserConfigMeta):

    @classmethod
    def read(cls):
        """Attempt to load the config from the default path, returning a new
        instance of the class.
        """
        return cls(addon_dir=Path(cls.__fields__['config_dir'].default,
                                  'addon_dir.txt').read_text(encoding='utf-8'))

    def write(self):
        """Write the active config to the default path."""
        (self.config_dir/'addon_dir.txt').write_text(str(self.addon_dir),
                                                     encoding='utf-8')




import asyncio
import bz2
from datetime import datetime as dt, timedelta
import re
import typing as T

from parsel import Selector
from yarl import URL

from .models import CacheEntry, Pkg, PkgOptions


EXPIRY = 3600


class BaseResolver:

    __members__ = {}

    synced: bool = False
    origin: str
    name: str

    def __init__(self, *, manager: 'Manager'):
        self.manager = manager
        self._sync_lock = asyncio.Lock(loop=manager.loop)

    def __init_subclass__(cls, origin: str, name: str):
        cls.__members__[origin] = cls
        cls.origin = origin
        cls.name = name

        orig_resolve = cls.resolve
        async def _resolve(self, *args, **kwargs):
            async with self._sync_lock:
                if not self.synced:
                    await self.sync()
                    self.synced = True
            return await orig_resolve(self, *args, **kwargs)
        cls.resolve = _resolve

    def __getattr__(self, name):
        # Delegate attribute access to the ``Manager``
        return getattr(self.manager, name)

    @classmethod
    def decompose_url(cls, url: str) -> T.Optional[T.Tuple[str, str]]:
        "Break a URL down into its component `origin` and `id`."

    async def sync(self) -> None:
        """Serves as a deferred, asynchronous `__init__`.  Do any
        preprocessing here if necessary, including writing to the cache.
        """

    async def resolve(self, id_or_slug: str, *,
                      strategy: str) -> Pkg:
        "Turn an ID or slug into a `models.Pkg`."
        raise NotImplementedError


class _CurseResolver(BaseResolver,
                     origin='curse', name='CurseForge'):

    _re_curse_url = re.compile(r'(?P<id>\d+)-(?P<slug>[a-z_-]+)')

    @classmethod
    def decompose_url(cls, url: str) -> T.Optional[T.Tuple[str, str]]:
        url = URL(url)
        if url.host in {'wow.curseforge.com', 'www.wowace.com'} \
                and len(url.parts) > 2 \
                and url.parts[1] == 'projects':
            return (cls.origin, url.parts[2])
        elif url.host == 'www.curseforge.com' \
                and len(url.parts) > 3 \
                and url.parts[1:3] == ('wow', 'addons'):
            slug = url.parts[3]
            match = cls._re_curse_url.match(slug)
            if match:
                slug = match.group('id')
            return (cls.origin, slug)

    async def resolve(self, id_or_slug: str, *,
                      strategy: str) -> Pkg:
        async with self.wc.get(f'https://wow.curseforge.com/projects/{id_or_slug}') \
                as response:
            if response.status != 200 \
                    or response.url.host not in {'wow.curseforge.com', 'www.wowace.com'}:
                raise self.PkgNonexistent
            content = await response.text()
        content = Selector(text=content, base_url=str(response.url))
        content.root.make_links_absolute()

        if strategy == 'canonical':
            file = content.css('.cf-recentfiles > li:first-child '
                               '.cf-recentfiles-credits-wrapper')
        else:
            file = max(map(int, content.css('.cf-recentfiles abbr::attr(data-epoch)')
                                       .extract()))
            file = content.xpath(f'//*[@class = "cf-recentfiles"]'
                                 f'//abbr[@data-epoch = "{file}"]/..')

        return Pkg(origin=self.origin,
                   id=content.xpath('''\
//div[@class = "info-label" and text() = "Project ID"]/following-sibling::div/text()'''
                                    ).extract_first().strip(),
                   slug=response.url.name,
                   name=content.css('meta[property="og:title"]::attr(content)')
                               .extract_first(),
                   description=content.css('meta[property="og:description"]::attr(content)')
                                      .extract_first(),
                   url=content.css('.view-on-curse > a::attr(href)').extract_first(),
                   file_id=URL(file.css('.overflow-tip::attr(href)')
                                   .extract_first()).name,
                   download_url=file.css('.fa-icon-download::attr(href)').extract_first(),
                   date_published=file.css('abbr::attr(data-epoch)').extract_first(),
                   version=file.css('.overflow-tip::attr(data-name)').extract_first(),
                   options=PkgOptions(strategy=strategy))


class _WowiResolver(BaseResolver,
                    origin='wowi', name='WoWInterface'):

    _data: dict

    _json_dump_url = 'https://api.mmoui.com/v3/game/WOW/filelist.json'
    _details_api_endpoint = 'https://api.mmoui.com/v3/game/WOW/filedetails/'

    _re_addon_url = re.compile(r'(?:download|info)(?P<slug>(?P<id>\d+)[^.]+)')

    @classmethod
    def decompose_url(cls, url: str) -> T.Optional[T.Tuple[str, str]]:
        url = URL(url)
        if url.host in {'wowinterface.com', 'www.wowinterface.com'} \
                and len(url.parts) == 3 \
                and url.parts[1] == 'downloads':
            match = cls._re_addon_url.match(url.name)
            if match:
                return (cls.origin, match.group('id'))

    async def sync(self) -> None:
        entry = self.db.query(CacheEntry).get((self.origin, self.origin))
        if not entry or ((entry.date_retrieved + timedelta(seconds=EXPIRY)) <
                         dt.now()):
            async with self.wc.get(self._json_dump_url) as resp:
                data = await resp.read()
            new_entry = CacheEntry(origin=self.origin, id=self.origin,
                                   date_retrieved=dt.now(), contents=data)
            entry = self.db.x_replace(new_entry, entry)

        self._data = {e['UID']: e for e in entry.contents}

    async def resolve(self, id_or_slug: str, *,
                      strategy: str) -> Pkg:
        file_id = id_or_slug.partition('-')[0]
        try:
            file = self._data[file_id]
        except KeyError:
            raise self.PkgNonexistent

        async with self.wc.get(f'{self._details_api_endpoint}/{file_id}.json') \
                as resp:
            details, = await resp.json()
        if file['UIDate'] != details['UIDate']:
            raise self.CacheObsolete

        return Pkg(origin=self.origin,
                   id=file['UID'],
                   slug=f'{file["UID"]}-{file["UIName"]}',
                   name=file['UIName'],
                   description=details['UIDescription'],
                   url=file['UIFileInfoURL'],
                   file_id=details['UIMD5'],
                   download_url=details['UIDownload'],
                   date_published=details['UIDate'],
                   version=details['UIVersion'],
                   options=PkgOptions(strategy=strategy))


class _TukuiResolver(BaseResolver,
                     origin='tukui', name='Tukui'):

    _re_addon_url = re.compile(re.escape('https://www.tukui.org/addons.php?id=') +
                               r'(?P<id>\d+)')

    @classmethod
    def decompose_url(cls, url: str) -> T.Optional[T.Tuple[str, str]]:
        match = cls._re_addon_url.match(url)
        if match:
            return (cls.origin, match.group('id'))

    async def resolve(self, id_or_slug: str, *,
                      strategy: str) -> Pkg:
        addon_id = id_or_slug.partition('-')[0]
        async with self.wc.get(f'https://www.tukui.org/api.php?addon={addon_id}') \
                as resp:
            if not resp.content_length:
                raise self.PkgNonexistent
            addon = await resp.json(content_type='text/html')

        return Pkg(origin=self.origin,
                   id=addon['id'],
                   slug=f'{addon["id"]}-{addon["name"]}',
                   name=addon['name'],
                   description=addon['small_desc'],
                   url=addon['web_url'],
                   file_id=addon['lastupdate'],
                   download_url=addon['url'],
                   date_published=addon['lastupdate'],
                   version=addon['version'],
                   options=PkgOptions(strategy=strategy))




from collections import namedtuple
import io
from pathlib import Path
import typing as T
import zipfile


class ExtractConflict(Exception):

    def __init__(self, conflicting_folders):
        super().__init__()
        self.conflicting_folders = conflicting_folders


class Archive:

    ExtractConflict = ExtractConflict

    def __init__(self, payload: bytes):
        self._archive = zipfile.ZipFile(io.BytesIO(payload))

    def extract(self, parent_folder: Path, *,
                overwrite: T.Union[bool, T.Set[str]]=False) -> None:
        "Extract the archive contents under ``parent_folder``."
        if overwrite is not True:
            conflicts = ({f.name for f in self.root_folders} &
                         ({f.name for f in parent_folder.iterdir()} -
                          (overwrite or set())))
            if conflicts:
                raise ExtractConflict(conflicts)
        self._archive.extractall(parent_folder)

    @property
    def root_folders(self) -> T.List[Path]:
        folders = sorted({Path(p).parts[0] for p in self._archive.namelist()})
        folders = [Path(p) for p in folders]
        return folders


_TocEntry = namedtuple('_TocEntry', 'key value')

class TocReader:
    """Extracts key–value pairs from TOC files."""

    def __init__(self, toc_file_path: Path):
        entries = (e.lstrip('# ').partition(': ')
                   for e in toc_file_path.read_text(encoding='utf-8').splitlines()
                   if e.startswith('## '))
        entries = {e[0]: e[2] for e in entries}
        self.entries = entries

    def __getitem__(self, keys: T.Union[str, T.Tuple[str]]) -> _TocEntry:
        if isinstance(keys, tuple):
            try:
                return next(filter(lambda i: i.value,
                                   (self.__getitem__(k) for k in keys)))
            except StopIteration:
                keys = keys[0]
        return _TocEntry(keys, self.entries.get(keys))




from collections import namedtuple
from functools import reduce
from textwrap import fill
import traceback
import typing as T
import webbrowser

import click
from outdated import check_outdated
from sqlalchemy import inspect
from texttable import Texttable

from . import __version__
from .config import UserConfig
from .manager import CliManager as Manager
from .models import Pkg, PkgFolder
from .utils import TocReader


_SUCCESS = click.style('✓', fg='green')
_FAILURE = click.style('✗', fg='red')

MESSAGES = {
    Manager.PkgInstalled:
        f'{_SUCCESS} {{0}}: installed {{1.new_pkg.version}}'
        f' from {{1.new_pkg.options.strategy}}'.format,
    Manager.PkgUpdated:
        f'{_SUCCESS} {{0}}: updated {{1.old_pkg.version}} to {{1.new_pkg.version}}'
        f' from {{1.new_pkg.options.strategy}}'.format,
    Manager.PkgRemoved:
        f'{_SUCCESS} {{}}: removed'.format,
    Manager.PkgAlreadyInstalled:
        f'{_FAILURE} {{}}: already installed'.format,
    Manager.PkgConflictsWithInstalled:
        lambda a, r: f'{_FAILURE} {a}: conflicts with installed '
                     f'add-on {_compose_addon_defn(r.conflicting_pkg)}',
    Manager.PkgConflictsWithPreexisting:
        f'{_FAILURE} {{}}: conflicts with an add-on not installed by instawow\n'
         '  pass `-o` to `install` if you do actually wish to overwrite this add-on'
        .format,
    Manager.PkgNonexistent:
        f'{_FAILURE} {{}}: no such project id or slug'.format,
    Manager.PkgNotInstalled:
        f'{_FAILURE} {{}}: not installed'.format,
    Manager.PkgOriginInvalid:
        f'{_FAILURE} {{}}: invalid origin'.format,
    Manager.PkgUpToDate:   # IGNORE
        ...,
    Manager.CacheObsolete:
        f'{_FAILURE} {{}}: the internal resolver cache is obsolete\n'
         '  run `instawow debug cache invalidate` and try again'
        .format,
    Manager.InternalError:
        lambda _, r: ''.join(traceback.format_exception(r.error.__class__,
                                                        r.error,
                                                        r.error.__traceback__)).rstrip(),}

_CONTEXT_SETTINGS = {'help_option_names': ['-h', '--help']}
_SEP = ':'

_parts = namedtuple('Parts', 'origin id_or_slug')


def _format_message(addon, result) -> str:
    return MESSAGES[result if isinstance(result, type) else result.__class__
                    ](addon, result)


def _tabulate(rows: T.List[T.Tuple[str, ...]], *,
              head: T.Tuple[str, ...]=(), show_index: bool=True) -> str:
    table = Texttable(max_width=0)
    table.set_deco(Texttable.HEADER | Texttable.VLINES)

    if show_index:
        table.set_cols_align(('r', *('l' for _ in rows[0])))
        head = ('', *head)
        rows = [(i, *v) for i, v in enumerate(rows, start=1)]
    table.add_rows([head, *rows])
    return table.draw()


def _compose_addon_defn(val):
    try:
        origin, slug = val.origin, val.slug
    except AttributeError:
        origin, slug = val
    return _SEP.join((origin, slug))


def _decompose_addon_defn(ctx, param, value):
    if isinstance(value, tuple):
        return [_decompose_addon_defn(ctx, param, v) for v in value]

    for resolver in ctx.obj.resolvers.values():
        parts = resolver.decompose_url(value)
        if parts:
            parts = _parts(*parts)
            break
    else:
        if _SEP not in value:
            raise click.BadParameter(value)
        parts = value.partition(_SEP)
        parts = _parts(parts[0], parts[-1])
    return _compose_addon_defn(parts), parts


def _init():
    addon_dir = UserConfig.default_addon_dir
    while True:
        try:
            UserConfig(addon_dir=addon_dir).create_dirs().write()
        except ValueError:
            if addon_dir:
                click.echo(f'{addon_dir!r} not found')
            addon_dir = click.prompt('Please enter the path to your add-on folder')
        else:
            break

init = click.Command(name='instawow-init', callback=_init,
                     context_settings=_CONTEXT_SETTINGS)


@click.group(context_settings=_CONTEXT_SETTINGS)
@click.version_option(__version__)
@click.option('--hide-progress', '-n', is_flag=True, default=False,
              help='Hide the progress bar.')
@click.pass_context
def main(ctx, hide_progress):
    """Add-on manager for World of Warcraft."""
    if not ctx.obj:
        while True:
            try:
                config = UserConfig.read()
            except (FileNotFoundError, ValueError):
                _init()
            else:
                break
        ctx.obj = manager = Manager(config=config,
                                    show_progress=not hide_progress)
        ctx.call_on_close(manager.close)

        is_outdated, _latest_version = check_outdated('instawow',
                                                      __version__)
        if is_outdated:
            click.echo(f'{click.style("!", fg="blue")} instawow is out of date:\n'
                       f'  run `python3 -m pip install -U instawow` to upgrade')

cli = main


@main.command()
@click.argument('addons', nargs=-1, required=True, callback=_decompose_addon_defn)
@click.option('--strategy', '-s',
              type=click.Choice(['canonical', 'latest']),
              default='canonical',
              help="Whether to install the latest published version "
                   "('canonical') or the very latest upload ('latest').")
@click.option('--overwrite', '-o',
              is_flag=True, default=False,
              help='Overwrite existing add-ons.')
@click.pass_obj
def install(manager, addons, overwrite, strategy):
    """Install add-ons."""
    for addon, result in zip((d for d, _ in addons),
                             manager.install_many((*p, strategy, overwrite)
                                                  for _, p in addons)):
        click.echo(_format_message(addon, result))


@main.command()
@click.argument('addons', nargs=-1, callback=_decompose_addon_defn)
@click.option('--strategy', '-s',
              type=click.Choice(['canonical', 'latest']), default=None,
              help="Whether to update to the latest published version "
                   "('canonical') or the very latest upload ('latest').")
@click.pass_obj
def update(manager, addons, strategy):
    """Update installed add-ons."""
    if not addons:
        addons = [(_compose_addon_defn(p), (p.origin, p.id))
                  for p in manager.db.query(Pkg).order_by(Pkg.slug).all()]
    for addon, result in zip((d for d, _ in addons),
                             manager.update_many((*p, strategy) for _, p in addons)):
        if not isinstance(result, Manager.PkgUpToDate):
            click.echo(_format_message(addon, result))


@main.command()
@click.argument('addons', nargs=-1, required=True, callback=_decompose_addon_defn)
@click.pass_obj
def remove(manager, addons):
    """Uninstall add-ons."""
    for addon, parts in addons:
        try:
            result = manager.remove(*parts)
        except (Manager.ManagerError, Manager.InternalError) as error:
            result = error
        click.echo(_format_message(addon, result))


@main.group('list')
def list_():
    """List add-ons."""


@list_.command()
@click.option('--column', '-c',
              multiple=True,
              help='A field to show in a column.  Nested fields are '
                   'dot-delimited.  Can be repeated.')
@click.option('--columns', '-C',
              is_flag=True, default=False,
              help='Print a list of all possible column values.')
@click.pass_obj
def installed(manager, column, columns):
    """List installed add-ons."""
    def _format_columns(pkg, columns):
        for column in columns:
            value = reduce(getattr, [pkg] + column.split('.'))
            if column == 'folders':
                value = '\n'.join(f.path.name for f in value)
            elif column == 'options':
                value = {'strategy': value.strategy}
            elif column == 'description':
                value = fill(value, width=50)
            yield value

    if columns:
        click.echo(_tabulate([(c,) for c in (*inspect(Pkg).columns.keys(),
                                             *inspect(Pkg).relationships.keys())],
                             head=('field',), show_index=False))
    else:
        pkgs = manager.db.query(Pkg).order_by(Pkg.origin, Pkg.slug).all()
        if not pkgs:
            return
        try:
            click.echo(_tabulate([(_compose_addon_defn(p),
                                   *_format_columns(p, column)) for p in pkgs],
                                 head=('add-on', *column)))
        except AttributeError as e:
            raise click.BadParameter(e.args)


@list_.command()
@click.pass_obj
def outdated(manager):
    """List outdated add-ons."""
    installed = manager.db.query(Pkg).order_by(Pkg.origin, Pkg.slug).all()
    new = manager.resolve_many((p.origin, p.id, p.options.strategy)
                               for p in installed)
    outdated = [(p, r) for p, r in zip(installed, new)
                if p.file_id != getattr(r, 'file_id', p.file_id)]
    if outdated:
        click.echo(_tabulate([(_compose_addon_defn(r),
                               p.version, r.version, r.options.strategy)
                              for p, r in outdated],
                             head=('add-on', 'current version',
                                   'new version', 'strategy')))


@list_.command()
@click.pass_obj
def preexisting(manager):
    """List add-ons not installed by instawow."""
    folders = {f.name
               for f in manager.config.addon_dir.iterdir() if f.is_dir()} - \
              {f.path.name for f in manager.db.query(PkgFolder).all()}
    folders = ((n, manager.config.addon_dir/n/f'{n}.toc') for n in folders)
    folders = {(n, TocReader(t)) for n, t in folders if t.exists()}
    if folders:
        click.echo(_tabulate([(n,
                               t['X-Curse-Project-ID'].value,
                               t['X-Curse-Packaged-Version', 'X-Packaged-Version',
                                 'Version'].value) for n, t in sorted(folders)],
                             head=('folder', 'curse id or slug', 'version')))


@main.command()
@click.argument('addon', callback=_decompose_addon_defn)
@click.pass_obj
def info(manager, addon):
    """Display installed add-on information."""
    pkg = addon[1]
    pkg = manager.db.x_unique(pkg.origin, pkg.id_or_slug)
    if pkg:
        rows = [('origin', pkg.origin),
                ('slug', pkg.slug),
                ('name', click.style(pkg.name, bold=True)),
                ('id', pkg.id),
                ('description', fill(pkg.description, max_lines=5)),
                ('homepage', click.style(pkg.url, underline=True)),
                ('version', pkg.version),
                ('release date', pkg.date_published),
                ('folders',
                 '\n'.join([str(pkg.folders[0].path.parent)] +
                           [' ├─ ' + f.path.name for f in pkg.folders[:-1]] +
                           [' └─ ' + pkg.folders[-1].path.name])),
                ('strategy', pkg.options.strategy),]
        click.echo(_tabulate(rows, show_index=False))
    else:
        click.echo(_format_message(addon[0], Manager.PkgNotInstalled))


@main.command()
@click.argument('addon', callback=_decompose_addon_defn)
@click.pass_obj
def hearth(manager, addon):
    """Open the add-on's homepage in your browser."""
    pkg = addon[1]
    pkg = manager.db.x_unique(pkg.origin, pkg.id_or_slug)
    if pkg:
        webbrowser.open(pkg.url)
    else:
        click.echo(_format_message(addon[0], Manager.PkgNotInstalled))


@main.command()
@click.argument('addon', callback=_decompose_addon_defn)
@click.pass_obj
def reveal(manager, addon):
    """Open the add-on folder in your file manager."""
    pkg = addon[1]
    pkg = manager.db.x_unique(pkg.origin, pkg.id_or_slug)
    if pkg:
        webbrowser.open(pkg.folders[0].path.as_uri())
    else:
        click.echo(_format_message(addon[0], Manager.PkgNotInstalled))


@main.group()
def debug():
    """Debugging funcionality."""


@debug.command()
@click.pass_context
def shell(ctx):
    """Drop into an interactive shell.

    The shell is created in context and provides access to the
    currently-active manager.
    """
    try:
        from IPython import embed
    except ImportError:
        click.echo('ipython is not installed')
    else:
        embed()


@debug.group()
def cache():
    """Manage the resolver cache."""


@cache.command()
@click.pass_obj
def invalidate(manager):
    """Invalidate the cache.  This sets the date retrieved for every
    cache entry to 1970 to trigger a recheck on the next run.
    """
    from datetime import datetime
    from .models import CacheEntry
    manager.db.query(CacheEntry).update({'date_retrieved': datetime.fromtimestamp(0)})
    manager.db.commit()


@cache.command()
@click.pass_obj
def clear(manager):
    """Nuke the cache from orbit."""
    from .models import CacheEntry
    manager.db.query(CacheEntry).delete()
    manager.db.commit()


@main.group()
def extras():
    """Additional functionality."""


@extras.group()
def bitbar():
    """Mini update GUI implemented as a BitBar plug-in.

    BitBar <https://getbitbar.com/> is a menu-bar app host for macOS.
    """


@bitbar.command()
@click.argument('caller')
@click.argument('version')
@click.pass_obj
def _generate(manager, caller, version):
    from base64 import b64encode
    from pathlib import Path
    from subprocess import run, PIPE

    outdated = manager.db.query(Pkg).order_by(Pkg.name).all()
    outdated = zip(outdated,
                   manager.resolve_many((p.origin, p.id, p.options.strategy)
                                        for p in outdated))
    outdated = [(p, r) for p, r in outdated
                if p.file_id != getattr(r, 'file_id', p.file_id)]

    icon = Path(__file__).parent/'assets'/f'''\
NSStatusItem-icon__\
{"has-updates" if outdated else "clear"}\
{"@2x" if b"Retina" in run(("system_profiler", "SPDisplaysDataType"),
                           stdout=PIPE).stdout else ""}\
.png'''
    icon = b64encode(icon.read_bytes()).decode()

    click.echo(f'''| templateImage="{icon}"
---
instawow ({__version__}:{version})''')
    if outdated:
        if len(outdated) > 1:
            click.echo(f'''\
Update all | bash={caller} param1=update terminal=false refresh=true''')
        for p, r in outdated:
            click.echo(f'''\
Update “{r.name}” ({p.version} ➞ {r.version}) | \
  refresh=true \
  bash={caller} param1=update param2={_compose_addon_defn(r)} terminal=false
View “{r.name}” on {manager.resolvers[r.origin].name} | \
  alternate=true \
  bash={caller} param1=hearth param2={_compose_addon_defn(r)} terminal=false''')


@bitbar.command()
def install():
    """Install the instawow BitBar plug-in."""
    from pathlib import Path
    import tempfile
    import sys

    with tempfile.TemporaryDirectory() as name:
        path = Path(name, 'instawow.1h.py')
        path.write_text(f'''\
#!/usr/bin/env LC_ALL=en_US.UTF-8 {sys.executable}

__version__ = {__version__!r}

import sys

from instawow.cli import main

main(['-n', *(sys.argv[1:] or ['extras', 'bitbar', '_generate', sys.argv[0], __version__])])
''')
        webbrowser.open(f'bitbar://openPlugin?src={path.as_uri()}')
        click.pause('Press any key to exit after installing the plug-in')
